; msLedStrip - microscope LED strip
; version 0.1 date 2009-06-18
; Copyright (C) 2009 Stefan Schuermans <stefan@blinkenarea.org>
; Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
; a BlinkenArea project - http://www.blinkenarea.org/

; clock frequency: 12 MHz (external crystal)

; PA0..1: crystal
; PA2: reset
; PB0: unused (output, low)
; PB1: unused (output, low)
; PB2: LED driver error input (TLC5943 XERR) (input, pull-up disabled)
; PB3: LED driver serial data loopback (TLC5943 SOUT) (input, pull-up disabled)
; PB4: LED driver blanking output (TLC5942 BLANK) (output, high)
; PB5: programming data input (input, pull-up enabled)
; PB6: programming data output (output, low)
; PB7: programming clock input (input, pull-up enabled)
; PD0: RS232 serial input (RX) (input, pull-up enabled)
; PD1: RS232 serial output (TX) (output, high)
; PD2: LED driver grayscale clock (TLC5943 GSCLK) (output, low)
; PD3: LED driver serial data (TLC5943 SIN) (output, low)
; PD4: LED driver data clock (TLC5943 SCLK) (output, low)
; PD5: LED driver brightness control select (TLC5943 BCSEL) (output, low)
; PD6: LED driver serial data latch (TC5943 XLAT) (output, low)

; serial protocol: 115200, 8N1
;   pixel data:
;     input:
;       MCUF <X>x<Y>-1/256, X=1..8, Y=7..12
;         message: <header> <data>
;         header: <magic> <height> <width> <channels> <maxval>
;         magic: 0x23 0x54 0x26 0x66
;         height: 0x00 0x07..0x0C (Y)
;         width: 0x00 0x01..0x08 (X)
;         channels: 0x00 0x01
;         maxval: 0x00 0xFF
;         data: <line0> ... <line<Y-1>>
;         line: <pixel0> ... <pixel<X-1>>
;         pixel: 0x00..0xFF
;     output:
;       MCUF <X-1>x<Y>-1/256
;         message: <header> <data>
;         header: <magic> <height> <width> <channels> <maxval>
;         magic: 0x23 0x54 0x26 0x66
;         height: 0x00 0x07..0x0C (Y)
;         width: 0x00 0x00..0x07 (X-1)
;         channels: 0x00 0x01
;         maxval: 0x00 0xFF
;         data: <line0> ... <line<Y-1>>
;         line: <pixel0> ... <pixel<X-2>>
;         pixel: 0x00..0xFF
;   PWM parameter data:
;     input:
;       message: <header> <data>
;       header: <magic> <y> <x>
;       magic: 0x50 0x41 0x52 0x4D
;       y: 0x00 0x00..0x0B (Y)
;       x: 0x00 0x00..0x07 (X)
;       data: <param 1> <param 2> <param 3> <param 4>
;       param I: byte, 0..255, coefficient for channel value ** I
;     output:
;       message: <header> <data>
;       header: <magic> <y> <x>
;       magic: 0x50 0x41 0x52 0x4D
;       y: 0x00 0x00..0x0B (Y)
;       x: 0x00 0x00..0x06,0xFF (X-1)
;       data: <param 1> <param 2> <param 3> <param 4>
;       param I: byte, 0..255, coefficient for channel value ** I
;   parameter save command:
;     input:
;       message: <header>
;       header: <magic>
;       magic: 0x53 0x41 0x56 0x45
;     output:
;       message: <header>
;       header: <magic>
;       magic: 0x53 0x41 0x56 0x45



.INCLUDE        "tn2313def.inc"



; constants
.equ    SER_IDLE_TICKS  =       250             ; number of ticks (20ms) serial port must be idle before animation is started



; IO pins
.equ    PORT_BLANK      =       PORTB           ; TLC5942 BLANK: PB4
.equ    BIT_BLANK       =       4
.equ    PORT_SIN        =       PORTD           ; TLC5943 SIN: PD3
.equ    BIT_SIN         =       3
.equ    PORT_SCLK       =       PORTD           ; TLC5943 SCLK: PD4
.equ    BIT_SCLK        =       4
.equ    PORT_BCSEL      =       PORTD           ; TLC5943 BCSEL: PD3
.equ    BIT_BCSEL       =       5
.equ    PORT_XLAT       =       PORTD           ; TLC5943 XLAT: PD4
.equ    BIT_XLAT        =       6



; general purpose registers
.def    DATA            =       r16
.def    TMP             =       r17
.def    CNT             =       r18
.def    CNT2            =       r19
.def    CNT3            =       r20
.def    CNT4            =       r21
.def    REG0            =       r0
.def    REG1            =       r1
.def    REG2            =       r2
.def    REG3            =       r3
.def    REG4            =       r4
.def    REG5            =       r5
.def    REG6            =       r6
.def    REG7            =       r7
.def    REG8            =       r8
.def    REG9            =       r9
.def    REG10           =       r10



; receive state
.equ    R_ST_WAIT       =       0               ; waiting for new message
.equ    R_ST_MCUF_HDR   =       1               ; receiving MCUF header
.equ    R_ST_MCUF_DATA  =       2               ; receiving MCUF data
.equ    R_ST_PARM_HDR   =       3               ; receiving PWM parameter header
.equ    R_ST_PARM_DATA  =       4               ; receiving PWM parameter data
.equ    R_ST_SAVE_HDR   =       5               ; receiving parameter save command
.def    RECV_STATE      =       r11

; receive position
;  - counts received data bytes of current frame
.def    RECV_POS        =       r12

; size of current MCUF frame being received
;  - only valid after MCUF header is received (set during reception of header)
.def    RECV_WIDTH      =       r13             ; width of current MCUF frame
.def    RECV_HEIGHT     =       r14             ; height of current MCUF frame

; X and Y position in current MCUF frame
;  - only valid during reception of MCUF data (not during reception of header)
;    and during reception of PARAM data (set during reception of header)
.def    RECV_POS_X      =       r22             ; X position in current MCUF frame
.def    RECV_POS_Y      =       r23             ; Y position in current MCUF frame

; number of ticks (20ms) serial port is idle
.def    SER_IDLE        =       r15

; position of paramerter saving
;  - address in EEPROM
;  - 0xFF if inactive
.def    PARAM_SAVE_POS  =       r24



.DSEG
.ORG    0x060

; current position in animation
ANIM_POS:       .BYTE   1

; current delay in animation in ticks (20ms)
ANIM_DELAY:     .BYTE   1

; receive buffer
;  - channel values (12 bytes)
;  - calibration position and data (4+8 bytes)
RECV_BUF:       .BYTE   12

; temporary buffer with channel values
TMP_BUF:        .BYTE   12

; PWM parameters
;  - for all pixels
;  - parameter format
;     <parameters> := <parameter 1> ... <parameter 4>
;     <parameter I> := byte, 0..255, coefficient for channel value ** I
PWM_PARAMS:     .BYTE   12*4



.MACRO brnex
        breq    EQ
        rjmp    @0
EQ:
.ENDM

.MACRO brlox
        brsh    SH
        rjmp    @0
SH:
.ENDM

.MACRO brshx
        brlo    LO
        rjmp    @0
LO:
.ENDM



.CSEG
.ORG    0x000
        rjmp    ENTRY                           ; RESET
        reti                                    ; INT0
        reti                                    ; INT1
        reti                                    ; TIMER1_CAPT
        reti                                    ; TIMER1_COMPA
        reti                                    ; TIMER1_OVF
        reti                                    ; TIMER0_OVF
        reti                                    ; USART0_RX
        reti                                    ; USART0_UDRE
        reti                                    ; USART0_TX
        reti                                    ; ANALOG_COMP
        reti                                    ; PC_INT
        reti                                    ; TIMER1_COMPB
        reti                                    ; TIMER0_COMPA
        reti                                    ; TIMER0_COMPB
        reti                                    ; USI_START
        reti                                    ; USI_OVERFLOW
        reti                                    ; EE_READY
        reti                                    ; WDT



; animation table
;  - contains animation to play when serial port idle
;  - format:
;      <entry> := <pixel 0> .. <pixel 11> <delay> <stuffing>
;      <pixel> := pixel value, byte, 0x00..0xFF
;      <delay> := delay in ticks (20ms), byte, 0x00..0xFF
;      <stuffing> := 3 zero bytes
ANIM_TAB:
.db     0x55,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0,0,0
.db     0xFF,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0,0,0
.db     0x00,0x55,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0,0,0
.db     0x00,0xFF,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0,0,0
.db     0x00,0x00,0x55,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0,0,0
.db     0x00,0x00,0xFF,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0,0,0
.db     0x00,0x00,0x00,0x55,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0,0,0
.db     0x00,0x00,0x00,0xFF,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0,0,0
.db     0x00,0x00,0x00,0x00,0x55,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0,0,0
.db     0x00,0x00,0x00,0x00,0xFF,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0,0,0
.db     0x00,0x00,0x00,0x00,0x00,0x55,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0,0,0
.db     0x00,0x00,0x00,0x00,0x00,0xFF,0x00,0x00,0x00,0x00,0x00,0x00, 50, 0,0,0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x55,0x00,0x00,0x00,0x00,0x00, 50, 0,0,0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0xFF,0x00,0x00,0x00,0x00,0x00, 50, 0,0,0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x55,0x00,0x00,0x00,0x00, 50, 0,0,0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xFF,0x00,0x00,0x00,0x00, 50, 0,0,0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x55,0x00,0x00,0x00, 50, 0,0,0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xFF,0x00,0x00,0x00, 50, 0,0,0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x55,0x00,0x00, 50, 0,0,0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xFF,0x00,0x00, 50, 0,0,0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x55,0x00, 50, 0,0,0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xFF,0x00, 50, 0,0,0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x55, 50, 0,0,0
.db     0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xFF, 50, 0,0,0
.db     0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55, 50, 0,0,0
.db     0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF, 50, 0,0,0
ANIM_TAB_END:



; code entry point
ENTRY:
; initialize stack pointer
        ldi     TMP,RAMEND
        out     SPL,TMP
; set clock prescaler to 1:1
        ldi     TMP,1<<CLKPCE
        out     CLKPR,TMP
        ldi     TMP,0
        out     CLKPR,TMP
; setup I/O pins
        ldi     TMP,0xB0                        ; PB[016]: output, low; PB4: output, high; PB[23]: input, pull-up off, PB[57]: input, pull-up on
        out     PORTB,TMP
        ldi     TMP,0x53
        out     DDRB,TMP
        ldi     TMP,0x03                        ; PD0: input, pull-up on; PD1: output, high;PD[2-6]: output, low
        out     PORTD,TMP
        ldi     TMP,0x7E
        out     DDRD,TMP
; disable analog comparator
        ldi     TMP,0x80
        out     ACSR,TMP
; set up timer 0 to 20ms interval (12MHz / 1024 / 234 = 50.08Hz)
        ldi     TMP,1<<WGM01                    ; clear timer on compare match
        out     TCCR0A,TMP
        ldi     TMP,1<<CS02|1<<CS00             ; prescaler 1:1024
        out     TCCR0B,TMP
        ldi     TMP,233                         ; count from 0 to 233
        out     OCR0A,TMP
; enable receiver of USART
        ldi     TMP,1<<U2X                      ; double speed
        out     UCSRA,TMP
        ldi     TMP,1<<RXEN|1<<TXEN             ; receiver enabled, transmitter enabled
        out     UCSRB,TMP
        ldi     TMP,1<<UCSZ1|1<<UCSZ0           ; asynchronous, 8N1: 8 data bits, no parity, 1 stop bit
        out     UCSRC,TMP
        ldi     TMP,0                           ; speed = Fosc / (8 * (UBRR + 1))
        out     UBRRH,TMP                       ;       = 12MHz / (8 * (12 + 1))
        ldi     TMP,12                          ;       = 115384bps (115200bps, 0.16% deviation)
        out     UBRRL,TMP
; enable watchdog (64ms)
        wdr
        ldi     TMP,1<<WDCE|1<<WDE
        out     WDTCR,TMP
        ldi     TMP,1<<WDE|1<<WDP1
        out     WDTCR,TMP
        wdr
; jump to main program
        rjmp    MAIN



; multiply a 16 bit value with an 8 bit value
; input: REG1:REG0 16 bit value, REG2 8 bit value
; output: REG5:REG4:REG3 24 bit result
; changes: TMP, CNT, REG0, REG1, REG2
; cycles: <= 107 (including rcall and ret)
MUL_16_8:
        clr     REG3                    ; result := 0
        clr     REG4
        clr     REG5
        clr     TMP                     ; upper part of first operand := 0
        ldi     CNT,8                   ; loop 8 times
MUL_16_8_LOOP:
        sbrs    REG2,0                  ; skip addition if bit is cleared
        rjmp    MUL_16_8_SKIP
        add     REG3,REG0               ; result += first operand
        adc     REG4,REG1
        adc     REG5,TMP
MUL_16_8_SKIP:
        lsl     REG0                    ; first operand <<= 1
        rol     REG1
        rol     TMP
        lsr     REG2                    ; second operand >>= 1
        dec     CNT                     ; bottom of loop
        brne    MUL_16_8_LOOP
        ret



; multiply a 16 bit value with an 8 bit value
; scale the result to 16 bit (so that FFFF * FF -> FFFF)
; function: f(x,y) = (x * y * (2^16+2^8+1) + 2^23) / 2^24
; input: REG1:REG0 16 bit value, REG2 8 bit value
; output: REG1:REG0 16 bit result
; changes: TMP, CNT, REG2, REG3, REG4, REG5
; cycles: <= 133 (including rcall and ret)
MUL_16_8_SCALE_16:
        rcall   MUL_16_8                ; intermed (REG5:REG4:REG3) := operand1 * operand2
        clr     TMP                     ; sum (REG1:REG0:REG2:CNT:TMP) := intermed << 16
        clr     CNT
        mov     REG2,REG3
        mov     REG0,REG4
        mov     REG1,REG5
        add     CNT,REG3                ; sum += intermed << 8
        adc     REG2,REG4
        adc     REG0,REG5
        adc     REG1,TMP                ;   TMP == 0 here
        add     TMP,REG3                ; sum += intermed
        adc     CNT,REG4
        adc     REG2,REG5
        clr     TMP                     ;   TMP := 0 (last 16 bit of sum no more needed)
        adc     REG0,TMP
        adc     REG1,TMP
        ldi     CNT,0x80                ; sum += 1 << 23 (last 16 bit of sum no more needed)
        add     REG2,CNT
        adc     REG0,TMP
        adc     REG1,TMP
        ret                             ; result (REG1:REG0) == sum >> 24



; get 16 bit PWM value from 8 bit channel value
; PWM calibration parameter format:
;   <parameters> := <parameter 1> ... <parameter 4>
;   <parameter I> := byte, 0..255, coefficient for channel value ** I
; input: REG6 8 bit channel value, X pointer to parameters
; output: REG8:REG7 16 bit PWM value, X pointer behind parameters
; changes: X, TMP, CNT, CNT2, REG0..REG10
; cycles: 982 (including rcall and ret)
CHANNEL2PWM:
        mov     REG0,REG6               ; sum (REG8:REG7) := parameter 1 * channel value ** 1
        mov     REG1,REG6
        ld      REG2,X+
        rcall   MUL_16_8_SCALE_16
        mov     REG7,REG0
        mov     REG8,REG1
        mov     REG9,REG6               ; channel value ** 1 (REG10:REG9) := channel value
        mov     REG10,REG6
        ldi     CNT2,3                  ; for N = 2 ... 4
CHANNEL2PWM_LOOP:
        mov     REG0,REG9               ; channel value ** N (REG10:REG9) := channel value ** (N - 1) * channel_value
        mov     REG1,REG10
        mov     REG2,REG6
        rcall   MUL_16_8_SCALE_16
        mov     REG9,REG0
        mov     REG10,REG1
        ld      REG2,X+                 ; sum += parameter N * channel value ** N
        rcall   MUL_16_8_SCALE_16
        add     REG7,REG0
        adc     REG8,REG1
        dec     CNT2                    ; bottom of loop
        brne    CHANNEL2PWM_LOOP
        ret



; shift out 8 bit value to PWM chip (MSB shifted first)
; input: REG0 value to shift out
; changes: CNT, REG0
PWM_SHIFT8:
; shift out byte
        ldi     CNT,8                           ; 8 bits
PWM_SHIFT8_LOOP:
        sbrs    REG0,7                          ; output bit to SIN
        cbi     PORT_SIN,BIT_SIN
        sbrc    REG0,7
        sbi     PORT_SIN,BIT_SIN
        lsl     REG0                            ; next bit
        sbi     PORT_SCLK,BIT_SCLK              ; output impulse on SCLK
        cbi     PORT_SCLK,BIT_SCLK
        dec     CNT                             ; loop
        brne    PWM_SHIFT8_LOOP
; done
        ret



; shift out 16 bit value to PWM chip (MSB shifted first)
; input: REG1:REG0 value to shift out
; changes: REG0, REG1, REG2
PWM_SHIFT16:
; save low byte
        mov     REG2,REG0
; shift out high byte
        mov     REG0,REG1
        rcall   PWM_SHIFT8
; shift out low byte
        mov     REG0,REG2
        rcall   PWM_SHIFT8
; done
        ret



; initialize PWM chip
; changes: TMP, CNT, CNT3, REG0, REG1, REG2
PWM_INIT:
; load brightness control register with 0x7F and turn on auto-repeat (0x80)
        sbi     PORT_BCSEL,BIT_BCSEL            ; switch to BC register
        ldi     TMP,0x80|0x7F                   ; shift out value 0x80|0x7F
        mov     REG0,TMP
        rcall   PWM_SHIFT8
        sbi     PORT_XLAT,BIT_XLAT              ; latch data into BC register
        cbi     PORT_XLAT,BIT_XLAT
; clear grayscale data registers
        cbi     PORT_BCSEL,BIT_BCSEL            ; switch to GS registers
        ldi     CNT3,16                         ; shift in 16 values
PWM_INIT_GS_LOOP:
        clr     REG0                            ;   shift out value 0
        clr     REG1
        rcall   PWM_SHIFT16
        dec     CNT3                            ; loop
        brne    PWM_INIT_GS_LOOP
        sbi     PORT_XLAT,BIT_XLAT              ; latch data into GS register
        cbi     PORT_XLAT,BIT_XLAT
; turn on PWM outputs
        cbi     PORT_BLANK,BIT_BLANK            ; clear BLANK
; done
        ret



; output new values in receive buffer to PWM chip
; input: Z pointer to channel values
; changes: X, Z, DATA, TMP, CNT, CNT2, CNT3, CNT4, REG0..REG10
PWM_OUTPUT:
; switch to GS registers
        cbi     PORT_BCSEL,BIT_BCSEL
; output 4 times 0 for unused PWM output
        clr     REG0
        clr     REG1
        rcall   PWM_SHIFT16
        rcall   PWM_SHIFT16
        rcall   PWM_SHIFT16
        rcall   PWM_SHIFT16
; output values for 12 pixels
        ldi     TMP,12                          ; move to position behind end of data buffer
        add     ZL,TMP
        clr     TMP
        adc     ZH,TMP
        ldi     CNT3,12                         ; head of pixel loop (12 pixels)
PWM_OUTPUT_LOOP_PIX:
        ldi     XL,low(PWM_PARAMS)              ; get pointer to PWM parameters of current pixel
        ldi     XH,high(PWM_PARAMS)
        mov     TMP,CNT3                        ;   offset = (CNT3 - 1) * 4
        dec     TMP
        lsl     TMP
        lsl     TMP
        add     XL,TMP                          ;   add offset
        clr     TMP
        adc     XH,TMP
        ld      DATA,-Z                         ; fetch value
        mov     REG6,DATA                       ; convert channel value to PWM value
        rcall   CHANNEL2PWM                     ; (in REG6: channel value, X parameters; out: REG8:REG7 PWM value)
        mov     REG0,REG7
        mov     REG1,REG8
        rcall   PWM_SHIFT16                     ; output value (input REG1:REG0)
        dec     CNT3                            ; bottom of pixel loop
        brne    PWM_OUTPUT_LOOP_PIX
; latch data into GS register
        sbi     PORT_XLAT,BIT_XLAT
        cbi     PORT_XLAT,BIT_XLAT
; done
        ret



; serial reception
; changes: X, Y, Z, DATA, TMP, CNT, CNT2, CNT3, CNT4, REG0..REG10
RECV:
; check for reception
        sbis    UCSRA,RXC
        rjmp    RECV_END                        ; nothing received
; check for errors
        sbic    UCSRA,FE
        rjmp    RECV_ERR                        ; frame error
        sbic    UCSRA,DOR
        rjmp    RECV_ERR                        ; data overrun
; read received byte from USART
        in      DATA,UDR
; check state
        ldi     TMP,RECV_TAB_END-RECV_TAB
        cp      RECV_STATE,TMP
        brlo    RECV_CHECKED                    ; state is valid -> use it
        ldi     TMP,R_ST_WAIT                   ; otherwise -> wait for first byte of header
        mov     RECV_STATE,TMP
RECV_CHECKED:
; jump according to state
        ldi     ZH,high(RECV_TAB)
        ldi     ZL,low(RECV_TAB)
        ldi     TMP,0
        add     ZL,RECV_STATE
        adc     ZH,TMP
        ijmp
RECV_TAB:
        rjmp    RECV_WAIT
        rjmp    RECV_MCUF_HDR
        rjmp    RECV_MCUF_DATA
        rjmp    RECV_PARM_HDR
        rjmp    RECV_PARM_DATA
        rjmp    RECV_SAVE_HDR
RECV_TAB_END:

; receive error
RECV_ERR:
        in      TMP,UDR                         ; read data from USART to clear error
        ldi     TMP,R_ST_WAIT                   ; not receiving a message, waiting for header
        mov     RECV_STATE,TMP
RECV_END:
        ret

; start waiting for first byte of header
RECV_DO_WAIT:
        ldi     TMP,R_ST_WAIT                   ; not receiving a message, waiting for header
        mov     RECV_STATE,TMP
        ret

; start waiting for first byte of header and check current byte
RECV_DO_WAIT_CHECK:
        ldi     TMP,R_ST_WAIT                   ; not receiving a message, waiting for header
        mov     RECV_STATE,TMP
        rjmp    RECV_WAIT

; waiting for header
RECV_WAIT:
        ldi     ZH,high(2*RECV_MCUF_HDR_DATA)   ; first byte of MCUF header
        ldi     ZL,low(2*RECV_MCUF_HDR_DATA)
        lpm     TMP,Z+                          ;   minimum
        cp      DATA,TMP
        brlo    RECV_WAIT_NO_MCUF
        lpm     TMP,Z                           ;   maximum
        cp      DATA,TMP
        brlo    RECV_WAIT_MCUF
        breq    RECV_WAIT_MCUF
RECV_WAIT_NO_MCUF:
        ldi     ZH,high(2*RECV_PARM_HDR_DATA)   ; first byte of calibration header
        ldi     ZL,low(2*RECV_PARM_HDR_DATA)
        lpm     TMP,Z+                          ;   minimum
        cp      DATA,TMP
        brlo    RECV_WAIT_NO_PARM
        lpm     TMP,Z                           ;   maximum
        cp      DATA,TMP
        brlo    RECV_WAIT_PARM
        breq    RECV_WAIT_PARM
RECV_WAIT_NO_PARM:
        ldi     ZH,high(2*RECV_SAVE_HDR_DATA)   ; first byte of parameter save command header
        ldi     ZL,low(2*RECV_SAVE_HDR_DATA)
        lpm     TMP,Z
        cp      DATA,TMP
        breq    RECV_WAIT_SAVE
        ret                                     ; not a first byte of any known header
; start receiving MCUF header
RECV_WAIT_MCUF:
        ldi     TMP,R_ST_MCUF_HDR               ; receive MCUF header
        mov     RECV_STATE,TMP
        clr     RECV_POS                        ; clear receive position
        rjmp    RECV_MCUF_HDR
; start receiving PWM parameter header
RECV_WAIT_PARM:
        ldi     TMP,R_ST_PARM_HDR               ; receive PWM parameter header
        mov     RECV_STATE,TMP
        clr     RECV_POS                        ; clear receive position
        rjmp    RECV_PARM_HDR
; start receiving parameter save command header
RECV_WAIT_SAVE:
        ldi     TMP,R_ST_SAVE_HDR               ; receive calibaration data header
        mov     RECV_STATE,TMP
        clr     RECV_POS                        ; clear receive position
        rjmp    RECV_SAVE_HDR

; expected MCUF header data (MCUF header 1..8x7..12-1/256)
;   - for every byte min and max is given
RECV_MCUF_HDR_DATA:
        .db     0x23, 0x23, 0x54, 0x54          ; magic
        .db     0x26, 0x26, 0x66, 0x66
        .db     0x00, 0x00, 0x07, 0x0C          ; height: 7..12
        .db     0x00, 0x00, 0x01, 0x08          ; width: 1..8
        .db     0x00, 0x00, 0x01, 0x01          ; channels: 1
        .db     0x00, 0x00, 0xFF, 0xFF          ; maxval: 255

; received a MCUF header byte
RECV_MCUF_HDR:
; check header byte
        ldi     ZH,high(2*RECV_MCUF_HDR_DATA)   ; load base address of MCUF header data ("2*" to convert word address to byte address)
        ldi     ZL,low(2*RECV_MCUF_HDR_DATA)
        clr     TMP                             ; add 2x offset (because table contains min/max pairs)
        add     ZL,RECV_POS
        adc     ZH,TMP
        add     ZL,RECV_POS
        adc     ZH,TMP
        lpm     REG0,Z+                         ; fetch minimum expected data byte from program memory
        lpm     REG1,Z                          ; fetch maximum expected data byte from program memory
        cp      DATA,REG0                       ; compare received data byte with minimum
        brlo    RECV_DO_WAIT_CHECK              ; too small -> start waiting for first byte of header and check current byte
        cp      DATA,REG1                       ; compare received data byte with maximum
        brlo    RECV_MCUF_HDR_OK
        breq    RECV_MCUF_HDR_OK
        rjmp    RECV_DO_WAIT_CHECK              ; too big -> start waiting for first byte of header and check current byte
RECV_MCUF_HDR_OK:
; save width and height
; also pass on header byte, but with decremented width
        mov     TMP,RECV_POS
        cpi     TMP,5                           ; position of height?
        brne    RECV_MCUF_HDR_NO_HEIGHT
        mov     RECV_HEIGHT,DATA                ; -> save height
RECV_MCUF_HDR_NO_HEIGHT:
        cpi     TMP,7                           ; position of width?
        brne    RECV_MCUF_HDR_NO_WIDTH
        mov     RECV_WIDTH,DATA                 ; -> save width
        dec     DATA                            ;    pass on decremented width
RECV_MCUF_HDR_NO_WIDTH:
        out     UDR,DATA                        ; pass on header byte
; next header byte
        inc     RECV_POS
        ldi     TMP,12                          ; check if entire header has been received
        cp      RECV_POS,TMP
        brsh    RECV_MCUF_HDR_END               ; entire header received -> start receiving data
        ret
; end of header: start receiving MCUF data
RECV_MCUF_HDR_END:
        ldi     TMP,R_ST_MCUF_DATA              ; receive MCUF data
        mov     RECV_STATE,TMP
        clr     RECV_POS                        ; clear receive position
        clr     RECV_POS_X
        clr     RECV_POS_Y
        ret

; received a MCUF data byte
RECV_MCUF_DATA:
; leftmost column?
        clr     TMP
        cp      RECV_POS_X,TMP
        breq    RECV_MCUF_DATA_LEFTMOST         ; leftmost -> jump to store
; pass on received data byte
        out     UDR,DATA
        rjmp    RECV_MCUF_DATA_NEXT
; store received byte in receive buffer
RECV_MCUF_DATA_LEFTMOST:
        ldi     ZH,high(RECV_BUF)               ; get base address of receive buffer
        ldi     ZL,low(RECV_BUF)
        clr     TMP                             ; get destination address (dest_addr := base_addr + pos_y)
        add     ZL,RECV_POS_Y
        adc     ZH,TMP
        st      Z,DATA                          ; store received byte at this address
; next position
RECV_MCUF_DATA_NEXT:
        inc     RECV_POS
        inc     RECV_POS_X
        cp      RECV_POS_X,RECV_WIDTH
        brlo    RECV_MCUF_DATA_NO_NEW_LINE
        clr     RECV_POS_X
        inc     RECV_POS_Y
RECV_MCUF_DATA_NO_NEW_LINE:
; check if frame is complete
        cp      RECV_POS_Y,RECV_HEIGHT
        brsh    RECV_MCUF_DATA_OUTPUT           ; end of frame -> output picture
        ret

; output picture
RECV_MCUF_DATA_OUTPUT:
        clr     SER_IDLE                        ; serial port is not idle
; fill unused rest of receive buffer with 0
        ldi     DATA,12                         ; until pos_y reaches 12
        ldi     ZH,high(RECV_BUF)               ; get base address of receive buffer
        ldi     ZL,low(RECV_BUF)
        clr     TMP                             ; get destination address (dest_addr := base_addr + pos_y)
        add     ZL,RECV_POS_Y
        adc     ZH,TMP
RECV_MCUF_DATA_OUTPUT_ZERO_LOOP:
        cp      RECV_POS_Y,DATA                 ; condition of loop
        brsh    RECV_MCUF_DATA_OUTPUT_ZERO_END
        st      Z+,TMP                          ; store zero
        inc     RECV_POS_Y                      ; next pixel
        rjmp    RECV_MCUF_DATA_OUTPUT_ZERO_LOOP ; bottom of loop
RECV_MCUF_DATA_OUTPUT_ZERO_END:
; output picture to PWM chip
        ldi     ZH,high(RECV_BUF)               ; get pointer to receive buffer containing new values
        ldi     ZL,low(RECV_BUF)
        rcall   PWM_OUTPUT                      ; output new values to PWM chip
; start waiting for next header
        rjmp    RECV_DO_WAIT

; expected PWM parameter header data
;   - for every byte min and max is given
RECV_PARM_HDR_DATA:
        .db     0x50, 0x50, 0x41, 0x41          ; magic
        .db     0x52, 0x52, 0x4D, 0x4D
        .db     0x00, 0x00, 0x00, 0x0B          ; y: 0..11
        .db     0x00, 0x00, 0x00, 0x07          ; x: 0..7

; received a PWM parameter header byte
RECV_PARM_HDR:
; check header byte
        ldi     ZH,high(2*RECV_PARM_HDR_DATA)   ; load base address of MCUF header data ("2*" to convert word address to byte address)
        ldi     ZL,low(2*RECV_PARM_HDR_DATA)
        clr     TMP                             ; add 2x offset (because table contains min/max pairs)
        add     ZL,RECV_POS
        adc     ZH,TMP
        add     ZL,RECV_POS
        adc     ZH,TMP
        lpm     REG0,Z+                         ; fetch minimum expected data byte from program memory
        lpm     REG1,Z                          ; fetch maximum expected data byte from program memory
        cp      DATA,REG0                       ; compare received data byte with minimum
        brlox   RECV_DO_WAIT_CHECK              ; too small -> start waiting for first byte of header and check current byte
        cp      DATA,REG1                       ; compare received data byte with maximum
        brlo    RECV_PARM_HDR_OK
        breq    RECV_PARM_HDR_OK
        rjmp    RECV_DO_WAIT_CHECK              ; too big -> start waiting for first byte of header and check current byte
RECV_PARM_HDR_OK:
; save X and Y coordinates
; also pass on header byte, but with decremented X coordinate
        mov     TMP,RECV_POS
        cpi     TMP,5                           ; position of Y coordinate?
        brne    RECV_PARM_HDR_NO_Y
        mov     RECV_POS_Y,DATA                 ; -> save Y coordinate
RECV_PARM_HDR_NO_Y:
        cpi     TMP,7                           ; position of X coordinate?
        brne    RECV_PARM_HDR_NO_X
        mov     RECV_POS_X,DATA                 ; -> save X coordinate
        dec     DATA                            ;    pass on decremented X coordinate
RECV_PARM_HDR_NO_X:
        out     UDR,DATA                        ; pass on header byte
; next header byte
        inc     RECV_POS
        ldi     TMP,8                           ; check if entire header has been received
        cp      RECV_POS,TMP
        brsh    RECV_PARM_HDR_END               ; entire header received -> start receiving position
        ret
; end of header: start receiving PWM parameters
RECV_PARM_HDR_END:
        ldi     TMP,R_ST_PARM_DATA              ; receive PWM parameter data
        mov     RECV_STATE,TMP
        clr     RECV_POS                        ; clear receive position
        ret

; received a PWM parameter data byte
RECV_PARM_DATA:
; X coordinate zero? (this means the PWM parameters are for this LED strip)
        clr     TMP
        cp      RECV_POS_X,TMP
        breq    RECV_PARM_DATA_STORE            ; X == 0 -> jump to store (do not pass on)
; pass on received data byte
        out     UDR,DATA
; move on
;  - storing also for X > 0 does not harm
;  - rest of processing (counting, detecting end of data) is needed also for X >= 0
; store received byte in receive buffer
RECV_PARM_DATA_STORE:
        ldi     ZH,high(RECV_BUF)               ; get base address of receive buffer
        ldi     ZL,low(RECV_BUF)
        clr     TMP                             ; get destination address (dest_addr := base_addr + pos)
        add     ZL,RECV_POS
        adc     ZH,TMP
        st      Z,DATA                          ; store received byte at this address
; next byte of PWM parameters
        inc     RECV_POS
; check if position and data is complete
        ldi     TMP,4
        cp      RECV_POS,TMP
        brsh    RECV_PARM_DATA_COMPLETE         ; data complete
        ret

; PWM parameter data complete
RECV_PARM_DATA_COMPLETE:
; X coordinate zero? (this means the PWM parameters are for this LED strip)
        clr     TMP
        cp      RECV_POS_X,TMP
        brne    RECV_PARM_DATA_COMPLETE_DONTUSE ; X != 0 -> do not use received parameters
; copy received PWM parameters into PWM parameter buffer
        ldi     ZH,high(RECV_BUF)               ; get base address of receive buffer
        ldi     ZL,low(RECV_BUF)
        ldi     YH,high(PWM_PARAMS)             ; get base address
        ldi     YL,low(PWM_PARAMS)
        mov     TMP,RECV_POS_Y                  ; offset := Y coordinate * 4
        lsl     TMP
        lsl     TMP
        add     YL,TMP                          ; destination := base + offset
        clr     TMP
        adc     YH,TMP
        ldi     CNT,4                           ; copy 4 bytes of data
RECV_PARM_DATA_COPY:
        ld      TMP,Z+
        st      Y+,TMP
        dec     CNT                             ; bottom of loop
        brne    RECV_PARM_DATA_COPY
RECV_PARM_DATA_COMPLETE_DONTUSE:
; start waiting for next header
        rjmp    RECV_DO_WAIT

; expected parameter save command data
RECV_SAVE_HDR_DATA:
        .db     0x53, 0x41, 0x56, 0x45          ; magic

; received a parameter save command header byte
RECV_SAVE_HDR:
; check header byte
        ldi     ZH,high(2*RECV_SAVE_HDR_DATA)   ; load base address of MCUF header data ("2*" to convert word address to byte address)
        ldi     ZL,low(2*RECV_SAVE_HDR_DATA)
        clr     TMP                             ; add offset
        add     ZL,RECV_POS
        adc     ZH,TMP
        lpm     REG0,Z                          ; fetch expected data byte from program memory
        cp      DATA,REG0                       ; compare with received data byte
        brnex   RECV_DO_WAIT_CHECK              ; mismatch -> start waiting for first byte of header and check current byte
; pass on header byte
        out     UDR,DATA
; next header byte
        inc     RECV_POS
        ldi     TMP,4                           ; check if entire header has been received
        cp      RECV_POS,TMP
        brsh    RECV_SAVE_HDR_END               ; entire header received -> start saving
        ret
; end of header: start saving
RECV_SAVE_HDR_END:
        clr     PARAM_SAVE_POS                  ; begin saving data
; start waiting for next header
        rjmp    RECV_DO_WAIT



; tick procedure - called every 20ms
; changes: X, Y, Z, DATA, TMP, CNT, CNT2, CNT3, CNT4, REG0..REG10
TICK:
; check if serial port is already idle long enough to run animation
        ldi     TMP,SER_IDLE_TICKS
        cp      SER_IDLE,TMP
        brsh    TICK_RUN_ANIM                   ; idle long enough -> run animation
; increment idle tick counter of serial port
        inc     SER_IDLE
; check serial port is now idle long enough to run animation
        ldi     TMP,SER_IDLE_TICKS
        cp      SER_IDLE,TMP
        brsh    TICK_START_ANIM                 ; idle long enough -> start animation
; done
        ret

; start animation
TICK_START_ANIM:
        clr     TMP
        sts     ANIM_POS,TMP                    ; set position in animation to begin
        sts     ANIM_DELAY,TMP                  ; no delay in animation active
; continue with running animation

; run animation
TICK_RUN_ANIM:
; check if delay is active
        lds     TMP,ANIM_DELAY
        cpi     TMP,0
        breq    TICK_RUN_ANIM_NO_DELAY
; decrement delay
        lds     TMP,ANIM_DELAY
        dec     TMP
        sts     ANIM_DELAY,TMP
; done
        ret

; run animation, no delay active
TICK_RUN_ANIM_NO_DELAY:
; get address of current position in animation table
        lds     ZH,ANIM_POS                     ; Z := ANIM_POS * 16
        swap    ZH
        mov     ZL,ZH
        cbr     ZL,0x0F
        cbr     ZH,0xF0
        ldi     TMP,low(2*ANIM_TAB)             ; Z := Z + ANIM_TAB
        add     ZL,TMP
        ldi     TMP,high(2*ANIM_TAB)
        adc     ZH,TMP
; get address of temporaray buffr
        ldi     YL,low(TMP_BUF)
        ldi     YH,high(TMP_BUF)
; read new values and new delay from animation table
        ldi     CNT,12                          ; read values for 12 pixels
TICK_RUN_ANIM_NO_DELAY_LOOP:
        lpm     TMP,Z+
        st      Y+,TMP
        dec     CNT
        brne    TICK_RUN_ANIM_NO_DELAY_LOOP
        lpm     TMP,Z+                          ; read delay
        sts     ANIM_DELAY,TMP
; next position in animation
        lds     TMP,ANIM_POS
        inc     TMP
        cpi     TMP,(2*ANIM_TAB_END - 2*ANIM_TAB) >> 4  ; number of entries in anim table
        brlo    TICK_RUN_ANIM_NO_END
        clr     TMP                             ; set position to begin if end of table reached
TICK_RUN_ANIM_NO_END:
        sts     ANIM_POS,TMP
; output picture to PWM chip
        ldi     ZH,high(TMP_BUF)                ; get pointer to receive buffer containing new values
        ldi     ZL,low(TMP_BUF)
        rcall   PWM_OUTPUT                      ; output new values to PWM chip
; done
        ret



; save parameters to EEPROM
; input: PARAM_SAVE_POS
; output: PARAM_SAVE_POS
; changes: TMP, DATA, CNT, X
PARAM_SAVE:
; exit if write to internal EEPROM is in progress
        sbic    EECR,EEWE
        ret
; check if a parameter byte shall be written
        cpi     PARAM_SAVE_POS,12*4
        brsh    PARAM_SAVE_NOT_PARAM
; get parameter byte
        ldi     XH,high(PWM_PARAMS)
        ldi     XL,low(PWM_PARAMS)
        clr     TMP
        add     XL,PARAM_SAVE_POS
        adc     XH,TMP
        ld      DATA,X
; start writing parameter byte
        clr     TMP
        out     EEARL,PARAM_SAVE_POS            ; put address into EEPROM address register
        out     EEDR,DATA                       ; put byte to write into EEPROM data register
        sbi     EECR,EEMWE                      ; initiate write
        sbi     EECR,EEWE
; increment write position
        inc     PARAM_SAVE_POS
; done
        ret
PARAM_SAVE_NOT_PARAM:
; check if a check byte shall be written
        cpi     PARAM_SAVE_POS,12*4+2
        brsh    PARAM_SAVE_NOT_CHECK
; calculate checksum
        ldi     XH,high(PWM_PARAMS)             ; get base address of parameters
        ldi     XL,low(PWM_PARAMS)
        clr     DATA
        ldi     CNT,12*4                        ; process every byte
PARAM_SAVE_CALC:
        ld      TMP,X+                          ; get next byte
        add     DATA,TMP                        ; one's complement sum
        clr     TMP
        adc     DATA,TMP
        dec     CNT                             ; bottom of loop
        brne    PARAM_SAVE_CALC
; complement byte if position is odd
        sbrc    PARAM_SAVE_POS,0
        com     DATA
; start writing check byte
        clr     TMP
        out     EEARL,PARAM_SAVE_POS            ; put address into EEPROM address register
        out     EEDR,DATA                       ; put byte to write into EEPROM data register
        sbi     EECR,EEMWE                      ; initiate write
        sbi     EECR,EEWE
; increment write position
        inc     PARAM_SAVE_POS
; done
        ret
PARAM_SAVE_NOT_CHECK:
; writing is not active
        ldi     PARAM_SAVE_POS,0xFF
; done
        ret



; main program - initialization
MAIN:
        wdr                                     ; watchdog reset
        rcall   PWM_INIT                        ; initialize PWM chip
        ldi     TMP,R_ST_WAIT                   ; not receiving a message, waiting for header
        mov     RECV_STATE,TMP
        ldi     TMP,SER_IDLE_TICKS              ; serial port is idle
        mov     SER_IDLE,TMP
        clr     TMP
        sts     ANIM_POS,TMP                    ; set position in animation to begin
        sts     ANIM_DELAY,TMP                  ; no delay in animation active
        ldi     PARAM_SAVE_POS,0xFF             ; not saving parameters

; load PWM parameters from EEPROM
        ldi     XH,high(PWM_PARAMS)             ; get address of buffer with PWM parameters
        ldi     XL,low(PWM_PARAMS)
        clr     DATA                            ; checksum
        clr     CNT2                            ; position
        ldi     CNT,12*4                        ; loop for 12 pixels with 4 bytes each
MAIN_PARAM_LOAD_LOOP:
        out     EEARL,CNT2                      ; read byte from EEPROM
        sbi     EECR,EERE
        in      TMP,EEDR
        st      X+,TMP                          ; store byte into parameter buffer
        add     DATA,TMP                        ; one's complement sum
        clr     TMP
        adc     DATA,TMP
        inc     CNT2                            ; next position
        dec     CNT                             ; bottom of loop
        brne    MAIN_PARAM_LOAD_LOOP
; check checksum
        out     EEARL,CNT2                      ; read checksum byte from EEPROM
        sbi     EECR,EERE
        in      TMP,EEDR
        cp      TMP,DATA
        brne    MAIN_PARAM_INIT                 ; mismatch -> initialize to default
        inc     CNT2                            ; next checksum byte
        com     DATA                            ; inverted
        out     EEARL,CNT2                      ; read checksum byte from EEPROM
        sbi     EECR,EERE
        in      TMP,EEDR
        cp      TMP,DATA
        brne    MAIN_PARAM_INIT                 ; mismatch -> initialize to default
        rjmp    MAIN_PARAM_INIT_END

; initialize PWM parameters to default values
MAIN_PARAM_INIT:
        ldi     XH,high(PWM_PARAMS)             ; get address of PWM parameter buffer
        ldi     XL,low(PWM_PARAMS)
        ldi     CNT,12                          ; loop for 12 pixels
MAIN_PARAM_INIT_LOOP:
        ldi     TMP,0x3F                        ; linear: 0.25
        st      X+,TMP
        ldi     TMP,0x00                        ; quadratic: 0.0
        st      X+,TMP
        ldi     TMP,0xC0                        ; cubic: 0.75
        st      X+,TMP
        ldi     TMP,0x00                        ; power of 4: 0.0
        st      X+,TMP
        dec     CNT                             ; bottom of loop
        brne    MAIN_PARAM_INIT_LOOP
MAIN_PARAM_INIT_END:



; main program - main loop
MAIN_LOOP:

; watchdog reset
        wdr

; receive serial data
        rcall   RECV

; check for compare match of timer 0
        in      TMP,TIFR
        sbrs    TMP,OCF0A
        rjmp    MAIN_NO_TICK
; call tick procedure
        rcall   TICK
; clear timer 0 compare match flag
        ldi     TMP,1<<OCF0A
        out     TIFR,TMP
MAIN_NO_TICK:

; save parameters to EEPROM
        rcall   PARAM_SAVE

; bottom of main loop
        rjmp    MAIN_LOOP

